@@ -24,13 +24,15 @@ my %lookup= (
     14 => [\&not_impl,   \&not_impl,        'TYPE_VARINT'],
     15 => [\&p2c_uuid,   \&c2p_uuid,        'TYPE_TIMEUUID'],
     16 => [\&not_impl,   \&not_impl,        'TYPE_INET'],
+    32 => [\&p2c_list,   \&c2p_list,        'TYPE_LIST'],
+    34 => [\&p2c_list,   \&c2p_list,        'TYPE_SET'],
 );
 
 sub not_impl { ... }
 sub _pack {
     my ($p, $l, $m, $i)= @_;
     $m //= '';
-    return "pack('l> $p', $l, (\$INPUT[$i] $m))";
+    return "pack('l> $p', $l, ($i $m))";
 }
 sub _unpack {
     my ($p, $l, $m, $v)= @_;
@@ -40,7 +42,7 @@ sub _unpack {
 
 sub p2c_string {
     my ($i)= @_;
-    return ("pack('l>/a', \$INPUT[$i])", "utf8::is_utf8(\$INPUT[$i]) && utf8::encode(\$INPUT[$i])");
+    return ("pack('l>/a', $i)", "utf8::is_utf8($i) && utf8::encode($i)");
 }
 sub c2p_string { return shift }
 sub c2p_utf8string { my $var= shift; return ($var, "utf8::decode $var") }
@@ -61,6 +63,42 @@ sub c2p_uuid { return _unpack('H[32]', 16, ' =~ s/\A(\w{8})(\w{4})(\w{4})(\w{4})
 #sub p2c_ { return   _pack('', , undef, @_) }
 #sub c2p_ { return _unpack('', , undef, @_) }
 
+sub p2c_list {
+    my ($i, $type)= @_;
+
+    my $t= $lookup{$type->[0]} or die "Unknown type $type->[0]";
+    my ($c, $prep)= $t->[0]('$copied_value', $type->[1]);
+    $prep //= '';
+
+    return
+        "pack('l>/a', (join '', pack('l>', 0+\@{$i}), map { my \$copied_value= \$_; $prep; defined \$copied_value ? ($c) : \$null } \@{$i}))",
+    ;
+}
+
+sub c2p_list {
+    my ($i, $type)= @_;
+
+    my $t= $lookup{$type->[0]} or die "Unknown type $type->[0]";
+    my ($c, $prep)= $t->[1]('$temp_val', $type->[1]);
+    $prep //= '';
+
+    return "do {
+                my \$rowcount= unpack('l>', substr(($i), 0, 4, ''));
+                my \@list;
+                for (1..\$rowcount) {
+                    my \$byte_count= unpack('l>', substr(($i), 0, 4, ''));
+                    if (\$byte_count >= 0) {
+                        my \$temp_val= substr(($i), 0, \$byte_count, '');
+                        $prep;
+                        push \@list, ($c);
+                    } else {
+                        push \@list, undef;
+                    }
+                }
+                \\\@list;
+            }";
+}
+
 our @EXPORT_OK= qw( build_row_encoder build_row_decoder );
 
 sub build_row_encoder {
@@ -75,30 +113,29 @@ sub build_row_encoder {
 
     my $result;
     for my $type (@$types) {
-        if (ref $type) { $type= $type->{type}; }
-        my $t= $lookup{$type} or die "Unknown type $type";
-        my ($c, $prep)= $t->[0]($i);
+        if (ref $type eq 'HASH') { $type= $type->{type}; }
+        my $t= $lookup{$type->[0]} or die "Unknown type $type->[0]";
+        my ($c, $prep)= $t->[0]("\$INPUT[$i]", $type->[1]);
 
         $code .= "    $prep if defined \$INPUT[$i];\n" if $prep;
         $result .= "        (defined \$INPUT[$i] ? ($c) : \$null) .\n";
         $i++;
     }
     $code = $code  . "    return\n        \$length_bits .\n" . substr($result, 0, -3). "\n    ;\n}";
-    return eval($code);
+    return(eval($code) or die $@);
 }
 
 sub build_row_decoder {
     my ($types)= @_;
-    my $count= 0+@$types;
 
     # $_ = [count, body, dest_rows]
     my $code= "sub {\n    local *OUTPUT= \$_[2];\n    my (\$byte_count, \$tmp_val);\n    for my \$row_id (1..\$_[0]) {\n        my \@row;\n";
 
     my $i= 0;
     for my $type (@$types) {
-        if (ref $type) { $type= $type->{type}; }
-        my $t= $lookup{$type} or die "Unknown type $type";
-        my ($c, $prep)= $t->[1]('$tmp_val');
+        if (ref $type eq 'HASH') { $type= $type->{type}; }
+        my $t= $lookup{$type->[0]} or die "Unknown type $type->[0]";
+        my ($c, $prep)= $t->[1]('$tmp_val', $type->[1]);
 
         $code .= '        $byte_count= unpack("l>", substr $_[1], 0, 4, "");'."\n";
         $code .= '        if ($byte_count >= 0) {'."\n";
@@ -110,7 +147,7 @@ sub build_row_decoder {
     }
 
     $code .= "        push \@OUTPUT, \\\@row;\n    }\n}";
-    return eval($code);
+    return(eval($code) or die $@);
 }
 
 1;
