@@ -23,7 +23,8 @@ sub execute {
 
     my $params= @bind_values ? \@bind_values : $sth->{cass_params};
     my $param_count = $sth->FETCH('NUM_OF_PARAMS');
-    return $sth->set_err($DBI::stderr, "Wrong number of parameters")
+
+    return $sth->set_err($DBI::stderr, sprintf "Wrong number of parameters. Expected %d, got %d", $param_count, 0+@$params)
         if @$params != $param_count;
 
     $sth->{cass_paging_state}= undef;
@@ -107,6 +108,11 @@ sub cass_read {
         my $decoder= $sth->{cass_row_decoder};
         my $rows_count= unpack('N', substr $body, 0, 4, '');
 
+        # No rows, no paging, that means we're done
+        if (!$rows_count && !$sth->{cass_paging_state}) {
+            $sth->STORE('Active', 0);
+        }
+
         $sth->{cass_row_decoder}->($rows_count, $body, ($data = []));
         1;
 
@@ -192,6 +198,10 @@ sub rows {
 sub DESTROY {
     my ($sth)= @_;
     finish_async($sth);
+
+    # This fixes an issue where DBI throws a warning for an 'insert into .. if not exists update ..',
+    # which (interestingly) returns rows
+    $sth->finish if $sth->FETCH('Active') && !$sth->FETCH('NUM_OF_FIELDS');
 }
 
 1;
